
/* Yahtzee, by John Kleinjans */
/*   including 100-pt bonuses and "wildcards" */
/*     for extra yahtzees */
/* Idea: convert into C++ and create a Dice class and a ScoreCard class */

#include <stdio.h>
#include <time.h>          /* time() */
#include <ctype.h>         /* toupper() */

typedef enum { false, true } bool;

/* -=- " random " functions -=- */
void seed_rand   ( void );
void init_scores ( int scores[] );
void init_scored ( bool scored[] );
void init_keep   ( bool keep[] );
int  getnum      ( void );

/* -=- dice functions -=- */
void roll_dice   ( int dice[], bool keep[], int scores[], bool scored[] );
void show_dice   ( int dice[] );
void choose_dice ( bool keep[] );

/* -=- scoring functions -=- */
void show_scores ( int scores[], bool scored[] );
int  choose_score( bool scored[] );
void score_it    ( int scores[], int dice[], int choice );
void set_sums    ( int dice[], int sums[] );

int k3( int sums[] );                      /* 3 of a kind */
int k4( int sums[] );                      /* 4 of a kind */
int fh( int sums[] );                      /* full house */
int ss( int sums[], bool yz_bonus );       /* short straight */
int ls( int sums[], bool yz_bonus );       /* long straight */
int yz( int sums[] );                      /* yahtzee */
int ch( int sums[] );                      /* chance */
int add_all_dice( int sums[] );            /* for k3(), k4(), and ch() */

bool yahtzee_bonus( int scores[], int sums[] );

/* -=- main() -=- */

int main( void )
{
    int  dice[6];     /* got 5 dice; face values 0x 1 2 3 4 5 */
    bool keep[6];     /* keep or roll again */
    int  scores[19];  /* includes totals & bonuses */
    bool scored[14];  /* use 1-13 */
    int  choice;      /* where to put the score */
    int playagain;
    int turn;
    char s[80];

    seed_rand();

    do
    {
        init_scores( scores );   /* set all scores to 0 */
        init_scored( scored );   /* set all to false */

        for ( turn=0; turn<13; turn++ )
        {
            roll_dice( dice, keep, scores, scored );
            choice = choose_score( scored );
            score_it( scores, dice, choice );
        }

        show_scores( scores, scored );
        printf("\nPlay again (y/n): ");
        fgets(s, 80, stdin);

    } while ( toupper( s[0] ) == 'Y' );

  return 0;
}

/* ----- FUNCTIONS ----- */

/***** initialization functions *****/

void seed_rand( void )
{
    time_t now;

    time( &now );
    srand( now );
}

void init_scores( int scores[] )
{
    int i;
    for ( i=0; i<19; i++ )
        scores[i] = 0;
}

void init_scored( bool scored[] )
{
    int i;
    for ( i=0; i<14; i++ )
        scored[i] = false;
}

void init_keep( bool keep[] )
{
    int i;
    for ( i=0; i<6; i++ )
        keep[i] = false;
}

int getnum( void ) /* Read an int from stdin */
{
   int i;
   char s[80];

   fgets(s, 80, stdin);
   if (s[0] == 'q') exit(0);
   sscanf(s, "%d", &i);
   return(i);
}

/***** dice functions *****/

void roll_dice( int dice[], bool keep[], int scores[], bool scored[] )
{
    int i;
    int roll;
    init_keep( keep );   /* set all to false */

    for ( roll=0; roll<3; roll++ )
    {
        for ( i=1; i<6; i++ )
            if ( keep[i] == false )
                dice[i] = rand()%6+1;

        show_scores( scores, scored );
        printf("Roll #%d:\n\n", roll+1);
        show_dice( dice );

        if ( roll<2 )
            choose_dice( keep );
    }
}

void show_dice( int dice[] )
{
    int i;

    printf("    ");
    for ( i=1; i<6; i++ )
        printf(" [%d]", dice[i]);
    printf("\n");

    printf("      1   2   3   4   5\n");

}

void choose_dice( bool keep[] )
{
    int choices;

    init_keep( keep );

    printf("\nDice to keep: (like 235) ");
    choices = getnum();
    printf("\n");

    while ( choices > 0 )
    {
        keep[ choices%10 ] = true;
        choices /= 10;
    }
}

/***** scoring functions *****/

void show_scores( int scores[], bool scored[] )
{
    int i;

    system("cls");
    printf("\n");
    printf(" 1  2  3  4  5  6  7  8  9 10 11 12 13  Totals  Bonus\n");
    printf("-- -- -- -- -- -- -- -- -- -- -- -- --  ------  -----\n");
    printf(" 1  2  3  4  5  6 3K 4K FH SS LS YZ CH  TT  TB  BT  BB TOT\n");

    for ( i=1; i<14; i++ )
    {
        if ( scores[i]<10 )
        {
            if ( scores[i] == 0 && scored[i] == true )
                printf("0");
            else printf(" ");
        }
        printf("%d ", scores[i]);
    }

    for ( i=14; i<19; i++ )
    {
        if ( scores[i]<100 ) printf(" ");
        if ( scores[i]<10  ) printf(" ");
        printf("%d ", scores[i]);
    }
    printf("\n\n");
}

int choose_score( bool scored[] )
{
    int choice;
    bool goodchoice = false;

    do
    {
        printf("\nScore as (1 - 13)? ");
        choice = getnum();

        if ( choice<1 || choice>13 )
            printf("Out of range. Choose again\n");
        else if ( scored[choice] == true )
            printf("%d taken. Choose again\n", choice);
        else
        {
            scored[choice] = true;
            goodchoice = true;
        }
    } while ( goodchoice == false );

    return choice;
}

void score_it( int scores[], int dice[], int c /*choice*/ )
{
    int sums[7];   /* dice faces 1-6 */
    int temp;
    bool yz_bonus = false;
    int i;

    set_sums( dice, sums );

    yz_bonus = yahtzee_bonus( scores, sums );

    if ( yz_bonus == true )
        scores[17] += 100;  /* do this *before* scoring the yahtzee */

    switch ( c )
    {
        case 1:             /* 1 through 5 fall through ... */
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:  scores[c] = c * sums[c]; break;
        case 7:  scores[c] = k3( sums );  break;            /* 3 of a kind */
        case 8:  scores[c] = k4( sums );  break;            /* 4 of a kind */
        case 9:  scores[c] = fh( sums );  break;            /* full house */
        case 10: scores[c] = ss( sums, yz_bonus );  break; /* small straight */
        case 11: scores[c] = ls( sums, yz_bonus );  break;  /* large straight */
        case 12: scores[c] = yz( sums );  break;            /* yahtzee */
        case 13: scores[c] = ch( sums );  break;            /* chance */
    }

/*" 1  2  3  4  5  6  7  8  9 10 11 12 13  totals   bonus\n"; */
/*" 1  2  3  4  5  6 3K 4K FH SS LS YZ CH  TT  TB  BT  BB  TOT\n"; */

    temp=0;
    for ( i=1; i<7; i++ )
        temp += scores[i];
    scores[14] = temp;              /* subtotal on top */

    if ( temp>62 ) scores[16] = 35; /* bonus on top */

    temp=0;                         /* bottom total */
    for ( i=7; i<14; i++ )
        temp += scores[i];
    scores[15]=temp;

    temp=0;                         /* grand total */
    for ( i=14; i<18; i++ )
        temp += scores[i];
    scores[18]=temp;
}

void set_sums( int dice[], int sums[] )
{
    int i;

    for ( i=1; i<7; i++ )
        sums[i]=0;

    for ( i=1; i<6; i++ )
        sums[ dice[i] ]++;
}

/***** the different hands *****/

int k3( int sums[] )                         /* 3 of a kind */
{
    bool is3k = false;
    int i;

    for ( i=1; i<7; i++ )
        if ( sums[i]>2 )
            is3k = true;

    if ( is3k == true )
        return add_all_dice( sums );
    else
        return 0;
}

int k4( int sums[] )                         /* 4 of a kind */
{
    bool is4k = false;
    int i;

    for ( i=1; i<7; i++ )
        if ( sums[i]>3 )
            is4k = true;

    if ( is4k == true )
        return add_all_dice( sums );
    else
        return 0;
}

int fh( int sums[] )                         /* full house */
{
    bool is2k = false;
    bool is3k = false;
    int i;

    for ( i=1; i<7; i++ )
    {
        if ( sums[i]==2 )
            is2k = true;
        if ( sums[i]==3 )
            is3k = true;
        if ( sums[i]==5 )
           is2k=true, is3k=true;
    }

    if ( is2k == true && is3k == true )
        return 25;
    else
        return 0;
}

int ss( int s[] /*sums[]*/, bool yz_bonus )  /* short straight */
{
    bool isss = false;
    int i;

    for ( i=0; i<3; i++ )
        if ( s[1+i] && s[2+i] && s[3+i] && s[4+i] )
            isss = true;

    if ( yz_bonus == true )      /* wildcard!! */
        isss = true;

    if ( isss == true )
        return 30;
    else
        return 0;
}

int ls( int s[] /*sums[]*/, bool yz_bonus )  /* long straight */
{
    bool isls = false;  /* this is actually not needed, */
                        /*   but it keeps the logic consistent */

    if ( s[2] && s[3] && s[4] && s[5] && ( s[1] || s[6] ) )
        isls = true;

    if ( yz_bonus == true )      /* wildcard!! */
        isls = true;

    if ( isls == true )
        return 40;
    else
        return 0;
}

int yz( int sums[] )                         /* yahtzee */
{
    bool isyz = false;
    int i;

    for ( i=1; i<7; i++ )
        if ( sums[i]==5 )
            isyz = true;

    if ( isyz == true )
        return 50;
    else
        return 0;
}

int ch( int sums[] )                         /* chance */
{
    return add_all_dice( sums );
}

int add_all_dice( int sums[] )               /* used for 3K, 4K, and chance */
{
    int total = 0;
    int i;

    for ( i=1; i<7; i++ )
        total += i*sums[i];

    return total;
}

bool yahtzee_bonus( int scores[], int sums[] )
{
    bool is_yz = false;
    int i;

    for ( i=1; i<7; i++ )
        if ( sums[i]==5 )
            is_yz = true;

    if ( (scores[12] != 0 ) && is_yz )
         return true;
    else
        return false;
}

